//-----------------------------------------------
// FracNEON_dp_v8a_mt1 Benchmark by Michael Kübel
//-----------------------------------------------
// command line options:
//
//  ./FracNEON_dp_v8a_mt1 <number of threads to create> <number of benchmark loops>
//  => e.g.
//  ./FracNEON dp_v8a_mt1 8 100
//  => create 8 threads and loop benchmark 100 times
//
// if no options are specified, the number of available CPU cores will the used for the thread number
// and the number of loops will be calculated to achieve around 10 seconds of benchmark run time

#include <iostream>
#include <iomanip>
#include <cstdint>
#include <thread>
#include <pthread.h>
#include <SDL2/SDL.h>
#define WINDOW_WIDTH 600

using namespace std;

unsigned int NUM_THREADS = std::thread::hardware_concurrency(); //get number of available cores

uint32_t pixels[WINDOW_WIDTH*WINDOW_WIDTH] = {0}; //32 bit pixel array mapped to texture
uint64_t global_iterations = {0};                 //64 bit global counter address
uint32_t global_y_line     = {0};                 //32 bit global y_line address

extern "C" uint64_t CalcMandelbrotASM_FracNEON_dp_v8a_mt1(const uint32_t* p,const uint64_t* iterations,const uint32_t* y,const uint32_t y_start);

void *worker(void *thread_number_ptr)
{
	//Start each thread with the thread_number as the first y_line to be calculated
	//After that each thread itself will increment and check the global_y_line within the assembler code
	int thread_number = *((int*)thread_number_ptr);
	CalcMandelbrotASM_FracNEON_dp_v8a_mt1(pixels,&global_iterations,&global_y_line,thread_number);
	return NULL;
}

int main(int argc, char** argv)
{
    const char prg_info[] = "FracNEON_dp_v8a_mt1 V1.0 by Michael Kübel";

    cout<<prg_info<<"\n";
    cout<<"-----------------------------------------\n";

	if ((argc>1) && (argc<4)) {
		 NUM_THREADS = stoi(argv[1],NULL);
		 if (NUM_THREADS<1) {NUM_THREADS=1;}
		 cout<<"Number of CPU threads chosen by user: "<<NUM_THREADS<<"\n";}
	else {cout<<"Number of concurrent CPU threads supported: "<<NUM_THREADS<<"\n";}
	
    pthread_t threads[NUM_THREADS];
    int thread_number_storage[NUM_THREADS];
    unsigned int repeat_benchmark;

    SDL_Event event;
    SDL_Init(SDL_INIT_VIDEO);
    SDL_Window   *window   = SDL_CreateWindow(prg_info,SDL_WINDOWPOS_UNDEFINED,SDL_WINDOWPOS_UNDEFINED,WINDOW_WIDTH,WINDOW_WIDTH,0);
    SDL_Renderer *renderer = SDL_CreateRenderer(window,-1,0);
    SDL_Texture  *texture  = SDL_CreateTexture(renderer,SDL_PIXELFORMAT_ARGB8888,SDL_TEXTUREACCESS_STATIC,WINDOW_WIDTH,WINDOW_WIDTH);
    SDL_SetRenderDrawColor(renderer,0,0,0,0);
    SDL_RenderClear(renderer);
    SDL_RenderPresent(renderer);

    cout<<"Calculating a Mandelbrot of 600x600 pixels at 4096 iterations with double-precision floats\n";
    
	//Calibrate benchmark timing by running 20 repeats (...about 1.2 seconds on a RPi400, Cortex A72, 4 cores, 1.8 Ghz)
	//but do that only if no repeat number is set by user by the second command line argument
    if (argc!=3) {
		cout<<"Calibrate timing...\n";
		uint64_t calibrate_start = SDL_GetPerformanceCounter();
		for (unsigned int f=0;f<20;f++) {
			global_y_line = NUM_THREADS-1; // initialize global y_line for each repeat
			for (unsigned int i=0;i<NUM_THREADS;i++) {
				thread_number_storage[i] = i;
				pthread_create(&threads[i],NULL,worker,&thread_number_storage[i]);
			}
			for (unsigned int i=0;i<NUM_THREADS;i++) {pthread_join(threads[i],NULL);}
		}
		uint64_t calibrate_end = SDL_GetPerformanceCounter();
		cout<<"Done...\n";
	
		float calibration_timing =(calibrate_end-calibrate_start)/(float)SDL_GetPerformanceFrequency();
		repeat_benchmark=static_cast<int>(10/calibration_timing*20);
	}
	else
	{repeat_benchmark = stoi(argv[2],NULL);}
	
	if (repeat_benchmark<1) {repeat_benchmark=1;}

	//Start benchmark with calculated repeats to reach about 5 seconds of benchmark timing
    cout<<"Benchmarking for around 10 seconds while looping "<<repeat_benchmark<<" times...\n";
    global_iterations=0; // reset global iterations
	uint64_t benchmark_start = SDL_GetPerformanceCounter();
    for (unsigned int f=0;f<repeat_benchmark;f++) {
        global_y_line = NUM_THREADS-1; // initialize global y_line for each repeat
		for (unsigned int i=0;i<NUM_THREADS;i++) {
			thread_number_storage[i] = i;
			pthread_create(&threads[i],NULL,worker,&thread_number_storage[i]);
		}
		for (unsigned int i=0;i<NUM_THREADS;i++) {pthread_join(threads[i],NULL);}
    }
    uint64_t benchmark_end = SDL_GetPerformanceCounter();
    cout<<"Done...\n";

    float benchmark_timing =(benchmark_end-benchmark_start)/(float)SDL_GetPerformanceFrequency();
    SDL_UpdateTexture(texture,NULL,pixels,WINDOW_WIDTH*sizeof(uint32_t)); //NULL = entire texture
    SDL_RenderCopy(renderer,texture,NULL,NULL); //NULL = entire texture, NULL = entire destination
    SDL_RenderPresent(renderer);
    cout<<"Iterations............: "<<global_iterations<<"\n";
    cout<<"Time taken............: "<<fixed<<setprecision(3)<<benchmark_timing<<" s\n";
    cout<<"Speed absolute........: "<<fixed<<setprecision(3)<<(global_iterations/benchmark_timing/1000000)<<" Million Iterations/s\n";
    cout<<"Speed per thread/core.: "<<fixed<<setprecision(3)<<(global_iterations/benchmark_timing/1000000/NUM_THREADS)<<" Million Iterations/s\n";
    cout<<"Close graphics window to quit...\n";
    while (1) {if (SDL_PollEvent(&event) && event.type == SDL_QUIT) break;}
    SDL_DestroyTexture(texture);
    SDL_DestroyRenderer(renderer);
    SDL_DestroyWindow(window);
    SDL_Quit();

    return 0;
}
