//--------------------------------------------------
// FracNEON_sp_v8a_mt1_to Benchmark by Michael Kübel
//--------------------------------------------------
// command line options:
//
//  ./FracNEON_sp_v8a_mt1_to <number of threads to create> <number of benchmark loops>
//  => e.g.
//  ./FracNEON sp_v8a_mt1_to 8 100
//  => create 8 threads and loop benchmark 100 times
//
// if no options are specified, the number of available CPU cores will the used for the thread number
// and the number of loops will be calculated to achieve around 10 seconds of benchmark run time

#include <iostream>
#include <iomanip>
#include <cstdint>
#include <thread>
#include <pthread.h>
#include <chrono>
#define WINDOW_WIDTH 600

using namespace std;

unsigned int NUM_THREADS = std::thread::hardware_concurrency(); //get number of available cores

uint32_t pixels[WINDOW_WIDTH*WINDOW_WIDTH] = {0}; //32 bit pixel array mapped to texture
uint64_t global_iterations = {0};                 //64 bit global counter address
uint32_t global_y_line     = {0};                 //32 bit global y_line address

extern "C" uint64_t CalcMandelbrotASM_FracNEON_sp_v8a_mt1_to(const uint32_t* p,const uint64_t* iterations,const uint32_t* y,const uint32_t y_start);

void *worker(void *thread_number_ptr)
{
	//Start each thread with the thread_number as the first y_line to be calculated
	//After that each thread itself will increment and check the global_y_line within the assembler code
	int thread_number = *((int*)thread_number_ptr);
	CalcMandelbrotASM_FracNEON_sp_v8a_mt1_to(pixels,&global_iterations,&global_y_line,thread_number);
	return NULL;
}

int main(int argc, char** argv)
{
    const char prg_info[] = "FracNEON_sp_v8a_mt1_to V1.0 by Michael Kübel";

    cout<<prg_info<<"\n";
    cout<<"--------------------------------------------\n";

	if ((argc>1) && (argc<4)) {
		 NUM_THREADS = stoi(argv[1],NULL);
		 if (NUM_THREADS<1) {NUM_THREADS=1;}
		 cout<<"Number of CPU threads chosen by user: "<<NUM_THREADS<<"\n";}
	else {cout<<"Number of concurrent CPU threads supported: "<<NUM_THREADS<<"\n";}
	
    pthread_t threads[NUM_THREADS];
    int thread_number_storage[NUM_THREADS];
    unsigned int repeat_benchmark;

    cout<<"Calculating a Mandelbrot of 600x600 pixels at 4096 iterations with single-precision floats\n";
    
	//Calibrate benchmark timing by running 20 repeats (...about 1.2 seconds on a RPi400, Cortex A72, 4 cores, 1.8 Ghz)
	//but do that only if no repeat number is set by user by the second command line argument
    if (argc!=3) {
		cout<<"Calibrate timing...\n";
		auto start1 = std::chrono::system_clock::now();
		for (unsigned int f=0;f<20;f++) {
			global_y_line = NUM_THREADS-1; // initialize global y_line for each repeat
			for (unsigned int i=0;i<NUM_THREADS;i++) {
				thread_number_storage[i] = i;
				pthread_create(&threads[i],NULL,worker,&thread_number_storage[i]);
			}
			for (unsigned int i=0;i<NUM_THREADS;i++) {pthread_join(threads[i],NULL);}
		}
    auto end1 = std::chrono::system_clock::now();
    chrono::duration <double> calibration_timing = end1-start1;	
    cout<<"Done...\n";
	
    repeat_benchmark=static_cast<int>(10/calibration_timing.count()*20);
	}
	else
	{repeat_benchmark = stoi(argv[2],NULL);}
	
	if (repeat_benchmark<1) {repeat_benchmark=1;}

	//Start benchmark with calculated repeats to reach about 5 seconds of benchmark timing
    cout<<"Benchmarking for around 10 seconds while looping "<<repeat_benchmark<<" times...\n";
    global_iterations=0; // reset global iterations
 	auto start2 = std::chrono::system_clock::now();
    for (unsigned int f=0;f<repeat_benchmark;f++) {
        global_y_line = NUM_THREADS-1; // initialize global y_line for each repeat
		for (unsigned int i=0;i<NUM_THREADS;i++) {
			thread_number_storage[i] = i;
			pthread_create(&threads[i],NULL,worker,&thread_number_storage[i]);
		}
		for (unsigned int i=0;i<NUM_THREADS;i++) {pthread_join(threads[i],NULL);}
    }
    auto end2 = std::chrono::system_clock::now();
    chrono::duration <double> benchmark_timing = end2-start2;	
    cout<<"Done...\n";
    cout<<"Iterations............: "<<global_iterations<<"\n";
    cout<<"Time taken............: "<<fixed<<setprecision(3)<<benchmark_timing.count()<<" s\n";
    cout<<"Speed absolute........: "<<fixed<<setprecision(3)<<(global_iterations/benchmark_timing.count()/1000000)<<" Million Iterations/s\n";
    cout<<"Speed per thread/core.: "<<fixed<<setprecision(3)<<(global_iterations/benchmark_timing.count()/1000000/NUM_THREADS)<<" Million Iterations/s\n";
    return 0;
}
